/*
 * Copyright (C) 2008, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: time.c,v 1.33 2008/07/24 05:27:36 Exp $
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/serial_reg.h>
#include <linux/interrupt.h>
#include <asm/addrspace.h>
#include <asm/io.h>
#include <asm/time.h>

#include <typedefs.h>
#include <osl.h>
#include <bcmutils.h>
#include <bcmnvram.h>
#include <sbconfig.h>
#include <sbchipc.h>
#include <siutils.h>
#include <hndmips.h>
#include <mipsinc.h>
#include <hndcpu.h>
#include <bcmdevs.h>

#include <wps_led.h> 

/* Global SI handle */
extern si_t *bcm947xx_sih;
extern spinlock_t bcm947xx_sih_lock;

/* Convenience */
#define sih bcm947xx_sih
#define sih_lock bcm947xx_sih_lock
#define WATCHDOG_MIN	3000	/* milliseconds */
#define WATCHDOG_MAX	5592	/* milliseconds */

extern int panic_timeout;
static uint32 watchdog = 0;


#define LED_BLINK_RATE_NORMAL   50
#define LED_BLINK_RATE_QUICK    10

int wps_led_pattern = 0;
int wps_led_state = 0;

int is_wl_secu_mode = 0;
static int wps_led_is_on = 0;
static int wps_led_state_old = 1;

static si_t *gpio_sih;
static int wps_led_init(void)
{
    if (!(gpio_sih = si_kattach(SI_OSH))) 
    {
        printk("%s failed!\n", __FUNCTION__);
        return -ENODEV;
    }

    return 0;
}

static int gpio_led_on_off(int gpio_num, int on_off)
{
    si_gpioreserve(gpio_sih, 1 << WPS_LED_GPIO, GPIO_APP_PRIORITY);
    si_gpioouten(gpio_sih, 1 << WPS_LED_GPIO, 1 << WPS_LED_GPIO, GPIO_APP_PRIORITY);
    si_gpioout(gpio_sih, 1 << WPS_LED_GPIO, on_off << WPS_LED_GPIO, GPIO_APP_PRIORITY);
    wps_led_is_on = on_off;
    return 0;
}

static void quick_blink2(void)
{
    static int interrupt_count = -1;

    interrupt_count++;
    if (interrupt_count == LED_BLINK_RATE_QUICK * 2)
        interrupt_count = 0;
    
    if (interrupt_count == 0)
        gpio_led_on_off(1, 1);
    else if (interrupt_count == LED_BLINK_RATE_QUICK)
        gpio_led_on_off(1, 0);
}

static void quick_blink(void)
{
    //static int blink_interval = 3000; /* 30 seconds */
    static int blink_interval = 500; /* 5 seconds */
    static int interrupt_count = -1;

    blink_interval--;
    interrupt_count++;
    if (interrupt_count == LED_BLINK_RATE_QUICK * 2)
        interrupt_count = 0;
    
    if (interrupt_count == 0)
        gpio_led_on_off(1, 1);
    else if (interrupt_count == LED_BLINK_RATE_QUICK)
        gpio_led_on_off(1, 0);
        
    if ( blink_interval <= 0 )
    {
        //blink_interval = 3000;
        blink_interval = 500;
        wps_led_state = 0;
    }
}

static int normal_blink(void)
{
    static int interrupt_count = -1;

    interrupt_count++;
    if (interrupt_count == LED_BLINK_RATE_NORMAL * 2)
        interrupt_count = 0;
    
    if (interrupt_count == 0)
        gpio_led_on_off(1, 1);
    else if (interrupt_count == LED_BLINK_RATE_NORMAL)
        gpio_led_on_off(1, 0);
}
static int wps_ap_lockdown_blink(void)
{
    static int interrupt_count = -1;

    interrupt_count++;
    if (interrupt_count == LED_BLINK_RATE_QUICK * 10)
        interrupt_count = 0;
    
    if (interrupt_count == 0)
        gpio_led_on_off(WPS_LED_GPIO, 1);
    else if (interrupt_count == LED_BLINK_RATE_QUICK)
        gpio_led_on_off(WPS_LED_GPIO, 0);
}

void __init
bcm947xx_time_init(void)
{
	unsigned int hz;

	/*
	 * Use deterministic values for initial counter interrupt
	 * so that calibrate delay avoids encountering a counter wrap.
	 */
	write_c0_count(0);
	write_c0_compare(0xffff);

	if (!(hz = si_cpu_clock(sih)))
		hz = 100000000;

	printk("CPU: BCM%04x rev %d at %d MHz\n", sih->chip, sih->chiprev,
	       (hz + 500000) / 1000000);

	/* Set MIPS counter frequency for fixed_rate_gettimeoffset() */
	mips_counter_frequency = hz / 2;

	/* Set watchdog interval in ms */
	watchdog = simple_strtoul(nvram_safe_get("watchdog"), NULL, 0);

	/* Ensure at least WATCHDOG_MIN
	 * Values above WATCHDOG_MAX overflow counter
	 */
	if (watchdog > 0) {
		if (watchdog < WATCHDOG_MIN)
			watchdog = WATCHDOG_MIN;

		if (watchdog > WATCHDOG_MAX)
			watchdog = WATCHDOG_MAX;
	}

	/* Set panic timeout in seconds */
	panic_timeout = watchdog / 1000;

    /* Init WLAN LED */
    wps_led_init(); 
}

#ifdef CONFIG_HND_BMIPS3300_PROF
extern bool hndprofiling;
#ifdef CONFIG_MIPS64
typedef u_int64_t sbprof_pc;
#else
typedef u_int32_t sbprof_pc;
#endif
extern void sbprof_cpu_intr(sbprof_pc restartpc);
#endif	/* CONFIG_HND_BMIPS3300_PROF */

static void
bcm947xx_timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
#ifdef CONFIG_HND_BMIPS3300_PROF
	/*
	 * Are there any ExcCode or other mean(s) to determine what has caused
	 * the timer interrupt? For now simply stop the normal timer proc if
	 * count register is less than compare register.
	 */
	if (hndprofiling) {
		sbprof_cpu_intr(regs->cp0_epc +
		                ((regs->cp0_cause >> (CAUSEB_BD - 2)) & 4));
		if (read_c0_count() < read_c0_compare())
			return;
	}
#endif	/* CONFIG_HND_BMIPS3300_PROF */

	/* Generic MIPS timer code */
	timer_interrupt(irq, dev_id, regs);

	/* Set the watchdog timer to reset after the specified number of ms */
	if (watchdog > 0) {
		if (sih->chip == BCM5354_CHIP_ID)
			si_watchdog(sih, WATCHDOG_CLOCK_5354 / 1000 * watchdog);
        else
           
            /* seems wrong watchdog value in original code */
			/* si_watchdog(sih, WATCHDOG_CLOCK / 1000 * watchdog); */
			si_watchdog_ms(sih, watchdog);
           
	}

#ifdef	CONFIG_HWSIM
	(*((int *)0xa0000f1c))++;
#endif

   
    /* Blink LED depending of WPS status */
    if ( wps_led_state == 0 )
    {
        if (wps_led_state_old != 0)
            gpio_led_on_off(1, 0);

        if ((!is_wl_secu_mode) && wps_led_is_on)
            gpio_led_on_off(1, 0);

        if (is_wl_secu_mode && (!wps_led_is_on))
            gpio_led_on_off(1, 1);
    }
    else
    if (wps_led_state == 1)
    {
        normal_blink();
    }
    else
    if (wps_led_state == 2)
    {
        quick_blink();
    }
    else
    if (wps_led_state == 3)
    {
        quick_blink2();
    }
    
    if (wps_led_state == 4)
    {
        wps_ap_lockdown_blink();
    }
    wps_led_state_old = wps_led_state;
 


}

static struct irqaction bcm947xx_timer_irqaction = {
	bcm947xx_timer_interrupt,
	SA_INTERRUPT,
	0,
	"timer",
	NULL,
	NULL
};

void __init
bcm947xx_timer_setup(struct irqaction *irq)
{
	/* Enable the timer interrupt */
	setup_irq(7, &bcm947xx_timer_irqaction);
}
